"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("@dcfjs/common/noCaptureEnv");
const ah = require("@dcfjs/common/arrayHelpers");
const sf = require("@dcfjs/common/serializeFunction");
const sr = require("@dcfjs/common/storageRegistry");
const XXHash = require('js-xxhash');
const pack = require('@dcfjs/objpack');
const v8 = require('v8');
let DeprecationWarningPrinted = false;
const envDeprecatedMsg = 'DeprecationWarning: manual env passing is deprecated and will be removed in future. use `captureEnv` or `registerCaptureEnv` instead.\nSee also: https://dcf.gitbook.io/dcf/guide/serialized-function-since-2.0';
function envDeprecated() {
    if (!DeprecationWarningPrinted) {
        console.warn(envDeprecatedMsg);
        DeprecationWarningPrinted = true;
    }
}
function hashPartitionFunc(numPartitions) {
    const seed = ((Math.random() * 0xffffffff) | 0) >>> 0;
    return sf.captureEnv((data) => {
        return XXHash.hash(pack.encode(data), seed) % numPartitions;
    }, {
        numPartitions,
        seed,
        XXHash: sf.requireModule('js-xxhash'),
        pack: sf.requireModule('@dcfjs/objpack'),
    });
}
function realGroupWith(rdds, context, numPartitions) {
    const rddCount = rdds.length;
    return context
        .union(...rdds.map((v, i) => v.map(sf.captureEnv(([k, v]) => {
        const ret = [];
        for (let j = 0; j < rddCount; j++) {
            ret.push(j === i ? [v] : []);
        }
        return [k, ret];
    }, { rddCount, i }))))
        .reduceByKey((a, b) => {
        const ret = [];
        for (let i = 0; i < a.length; i++) {
            ret.push(a[i].concat(b[i]));
        }
        return ret;
    }, numPartitions);
}
function attachFinalizedFunc(func, finalized) {
    if (!finalized) {
        return func;
    }
    return sf.captureEnv(async (arg, ts) => {
        await finalized(ts);
        return func(arg);
    }, {
        func,
        finalized,
    });
}
class RDD {
    constructor(context) {
        this._context = context;
    }
    // Dependence rdd should override this method to get better performance.
    getNumPartitions() {
        const tmp = this.getFunc();
        if (tmp instanceof Promise) {
            return tmp.then((v) => v[0]);
        }
        return tmp[0];
    }
    union(...others) {
        return this._context.union(this, ...others);
    }
    async collect() {
        const [numPartitions, partitionFunc, finalizedFunc] = await this.getFunc();
        return this._context.execute(numPartitions, partitionFunc, attachFinalizedFunc(ah.concatArrays, finalizedFunc));
    }
    async count() {
        const [numPartitions, partitionFunc, finalizedFunc] = await this.getFunc();
        return this._context.execute(numPartitions, sf.captureEnv((partitionId) => {
            const f = partitionFunc(partitionId);
            return sf.captureEnv((workerId) => {
                return Promise.resolve(f(workerId)).then((v) => v.length);
            }, { f });
        }, {
            partitionFunc,
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        }), attachFinalizedFunc((v) => v.reduce((a, b) => a + b, 0), finalizedFunc));
    }
    async take(count) {
        const [numPartitions, partitionFunc, finalizedFunc] = await this.getFunc();
        return this._context.execute(numPartitions, sf.captureEnv((partitionId) => {
            const f = partitionFunc(partitionId);
            return sf.captureEnv((workerId) => {
                return Promise.resolve(f(workerId)).then((v) => v.slice(0, count));
            }, { f, count });
        }, {
            count,
            partitionFunc,
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        }), attachFinalizedFunc(sf.captureEnv((v) => ah.concatArrays(v).slice(0, count), {
            count,
            ah: sf.requireModule('@dcfjs/common/arrayHelpers'),
        }), finalizedFunc));
    }
    max(comparator = ah.defaultComparator) {
        return this.reduce(sf.captureEnv((a, b) => {
            return comparator(a, b) > 0 ? a : b;
        }, {
            comparator,
        }));
    }
    min(comparator = ah.defaultComparator) {
        return this.reduce(sf.captureEnv((a, b) => {
            return comparator(a, b) < 0 ? a : b;
        }, {
            comparator,
        }));
    }
    mapPartitions(transformer, env) {
        if (env) {
            sf.captureEnv(transformer, env);
            envDeprecated();
        }
        return new MappedRDD(this._context, this, transformer);
    }
    glom() {
        return this.mapPartitions((v) => [v]);
    }
    map(transformer, env) {
        if (env) {
            sf.captureEnv(transformer, env);
            envDeprecated();
        }
        return this.mapPartitions(sf.captureEnv((v) => v.map((v) => transformer(v)), { transformer }));
    }
    flatMap(transformer, env) {
        if (env) {
            sf.captureEnv(transformer, env);
            envDeprecated();
        }
        return this.mapPartitions(sf.captureEnv((v) => ah.concatArrays(v.map((v) => transformer(v))), {
            transformer,
            ah: sf.requireModule('@dcfjs/common/arrayHelpers'),
        }));
    }
    filter(filterFunc, env) {
        if (env) {
            sf.captureEnv(filterFunc, env);
            envDeprecated();
        }
        return this.mapPartitions(sf.captureEnv((v) => v.filter((v) => filterFunc(v)), { filterFunc }));
    }
    async saveAsTextFile(baseUrl, { overwrite = true, encoding = 'utf8', extension = 'txt', compressor, functionEnv, } = {}) {
        const [numPartitions, partitionFunc, finalizedFunc] = await this.getFunc();
        if (functionEnv && compressor) {
            sf.captureEnv(compressor, functionEnv);
            envDeprecated();
        }
        let selectedLoader = this._context.getFileLoader(baseUrl);
        if (!selectedLoader) {
            throw new Error('No loader found for ' + baseUrl);
        }
        await selectedLoader.initSaveProgress(baseUrl, overwrite);
        const saveFileFunc = selectedLoader.saveFile;
        const markSaveSuccessFunc = selectedLoader.markSaveSuccess;
        return this._context.execute(numPartitions, sf.captureEnv((partitionId) => {
            const f = partitionFunc(partitionId);
            return sf.captureEnv((workerId) => {
                return Promise.resolve(f(workerId))
                    .then((arr) => {
                    let content = Buffer.from(arr.join('\n'));
                    return compressor
                        ? Promise.resolve(compressor(content))
                        : Promise.resolve(content);
                })
                    .then((content) => {
                    return saveFileFunc(baseUrl, `part-${partitionId}.${extension}`, content);
                });
            }, {
                f,
                baseUrl,
                compressor,
                encoding,
                extension,
                partitionId,
                saveFileFunc,
                sf: sf.requireModule('@dcfjs/common/serializeFunction'),
            });
        }, {
            baseUrl,
            compressor,
            encoding,
            extension,
            partitionFunc,
            saveFileFunc,
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        }), attachFinalizedFunc(sf.captureEnv(async () => {
            await markSaveSuccessFunc(baseUrl);
        }, {
            baseUrl,
            markSaveSuccessFunc,
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        })));
    }
    async reduce(reduceFunc, env) {
        const [numPartitions, partitionFunc, finalizedFunc] = await this.getFunc();
        if (env) {
            sf.captureEnv(reduceFunc, env);
            envDeprecated();
        }
        return this._context.execute(numPartitions, sf.captureEnv((partitionId) => {
            const f = partitionFunc(partitionId);
            return sf.captureEnv((workerId) => {
                return Promise.resolve(f(workerId)).then((arr) => {
                    let lastRes = arr[0];
                    for (let i = 1; i < arr.length; i++) {
                        lastRes = reduceFunc(lastRes, arr[i]);
                    }
                    return lastRes;
                });
            }, { f, reduceFunc });
        }, {
            partitionFunc,
            reduceFunc,
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        }), attachFinalizedFunc(sf.captureEnv((arr) => {
            arr = arr.filter((v) => v !== undefined);
            let lastRes = arr[0];
            for (let i = 1; i < arr.length; i++) {
                lastRes = reduceFunc(lastRes, arr[i]);
            }
            return lastRes;
        }, {
            reduceFunc,
        }), finalizedFunc));
    }
    persist(storageType) {
        return new CachedRDD(this._context, this, storageType || this._context.option.defaultPersistStorage);
    }
    // cache is only a alias of persist in dcf 2.0
    cache(storageType) {
        return this.persist(storageType);
    }
    partitionBy(numPartitions, partitionFunc, env) {
        if (env) {
            sf.captureEnv(partitionFunc, env);
            envDeprecated();
        }
        return new RepartitionRDD(this._context, this, numPartitions, sf.captureEnv((partitionId) => sf.captureEnv((data) => {
            const regrouped = [];
            for (let i = 0; i < numPartitions; i++) {
                regrouped[i] = [];
            }
            for (const item of data) {
                const parId = partitionFunc(item);
                regrouped[parId].push(item);
            }
            return regrouped;
        }, { numPartitions, partitionFunc }), {
            numPartitions,
            partitionFunc,
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        }));
    }
    coalesce(numPartitions) {
        return new CoalesceRDD(this._context, this, numPartitions);
    }
    repartition(numPartitions) {
        return this.partitionBy(numPartitions, sf.captureEnv(() => (Math.random() * numPartitions) | 0, {
            numPartitions,
        }));
    }
    distinct(numPartitions = this._context.option.defaultPartitions) {
        return this.partitionBy(numPartitions, hashPartitionFunc(numPartitions)).mapPartitions(sf.captureEnv((datas) => {
            const ret = [];
            const map = {};
            for (const item of datas) {
                const k = pack.encode(item).toString('base64');
                if (!map[k]) {
                    map[k] = item;
                    ret.push(item);
                }
            }
            return ret;
        }, {
            pack: sf.requireModule('@dcfjs/objpack'),
        }));
    }
    combineByKey(createCombiner, mergeValue, mergeCombiners, numPartitions = this._context.option.defaultPartitions, partitionFunc = hashPartitionFunc(numPartitions), env) {
        if (env) {
            sf.captureEnv(createCombiner, env);
            sf.captureEnv(mergeValue, env);
            sf.captureEnv(mergeCombiners, env);
            sf.captureEnv(partitionFunc, env);
            envDeprecated();
        }
        const mapFunction1 = sf.captureEnv((datas) => {
            const ret = [];
            const map = {};
            for (const item of datas) {
                const k = pack.encode(item[0]).toString('base64');
                let r = map[k];
                if (!r) {
                    r = [item[0], createCombiner(item[1])];
                    map[k] = r;
                    ret.push(r);
                }
                else {
                    r[1] = mergeValue(r[1], item[1]);
                }
            }
            return ret;
        }, {
            createCombiner,
            mergeValue,
            pack: sf.requireModule('@dcfjs/objpack'),
        });
        const mapFunction2 = sf.captureEnv((datas) => {
            const ret = [];
            const map = {};
            for (const item of datas) {
                const k = pack.encode(item[0]).toString('base64');
                let r = map[k];
                if (!r) {
                    r = [item[0], item[1]];
                    map[k] = r;
                    ret.push(r);
                }
                else {
                    r[1] = mergeCombiners(r[1], item[1]);
                }
            }
            return ret;
        }, {
            mergeCombiners,
            pack: sf.requireModule('@dcfjs/objpack'),
        });
        const realPartitionFunc = sf.captureEnv((data) => {
            return partitionFunc(data[0]);
        }, {
            partitionFunc,
        });
        return this.mapPartitions(mapFunction1)
            .partitionBy(numPartitions, realPartitionFunc)
            .mapPartitions(mapFunction2);
    }
    reduceByKey(func, numPartitions = this._context.option.defaultPartitions, partitionFunc, env) {
        return this.combineByKey((x) => x, func, func, numPartitions, partitionFunc, env);
    }
    groupWith(...others) {
        return realGroupWith([this, ...others], this._context);
    }
    cogroup(other, numPartitions) {
        return realGroupWith([this, other], this._context, numPartitions);
    }
    join(other, numPartitions) {
        return this.cogroup(other, numPartitions).flatMap(([k, [v1s, v2s]]) => {
            const ret = [];
            for (const v1 of v1s) {
                for (const v2 of v2s) {
                    ret.push([k, [v1, v2]]);
                }
            }
            return ret;
        });
    }
    leftOuterJoin(other, numPartitions) {
        return this.cogroup(other, numPartitions).flatMap(([k, [v1s, v2s]]) => {
            const ret = [];
            if (v2s.length === 0) {
                for (const v1 of v1s) {
                    ret.push([k, [v1, null]]);
                }
            }
            else {
                for (const v1 of v1s) {
                    for (const v2 of v2s) {
                        ret.push([k, [v1, v2]]);
                    }
                }
            }
            return ret;
        });
    }
    rightOuterJoin(other, numPartitions) {
        return this.cogroup(other, numPartitions).flatMap(([k, [v1s, v2s]]) => {
            const ret = [];
            if (v1s.length === 0) {
                for (const v2 of v2s) {
                    ret.push([k, [null, v2]]);
                }
            }
            else {
                for (const v1 of v1s) {
                    for (const v2 of v2s) {
                        ret.push([k, [v1, v2]]);
                    }
                }
            }
            return ret;
        });
    }
    fullOuterJoin(other, numPartitions) {
        return this.cogroup(other, numPartitions).flatMap(([k, [v1s, v2s]]) => {
            const ret = [];
            if (v1s.length === 0) {
                for (const v2 of v2s) {
                    ret.push([k, [null, v2]]);
                }
            }
            else if (v2s.length === 0) {
                for (const v1 of v1s) {
                    ret.push([k, [v1, null]]);
                }
            }
            else {
                for (const v1 of v1s) {
                    for (const v2 of v2s) {
                        ret.push([k, [v1, v2]]);
                    }
                }
            }
            return ret;
        });
    }
    sort(ascending = true, numPartitions) {
        return this.sortBy((v) => v, ascending, numPartitions);
    }
    sortBy(keyFunc, ascending = true, numPartitions, env) {
        return new SortedRDD(this._context, this, numPartitions || this._context.option.defaultPartitions, keyFunc, ascending);
    }
}
exports.RDD = RDD;
class GeneratedRDD extends RDD {
    constructor(context, numPartition, func) {
        super(context);
        this._numPartition = numPartition;
        this._function = func;
    }
    getFunc() {
        return [this._numPartition, this._function, undefined];
    }
    getNumPartitions() {
        return this._numPartition;
    }
}
exports.GeneratedRDD = GeneratedRDD;
class MappedRDD extends RDD {
    constructor(context, dependence, mapper) {
        super(context);
        this._mapper = mapper;
        this._dependence = dependence;
    }
    async getFunc() {
        const mapper = this._mapper;
        const [numPartitions, depFunc, depFinalizer,] = await this._dependence.getFunc();
        return [
            numPartitions,
            sf.captureEnv((partitionId) => {
                const f = depFunc(partitionId);
                return sf.captureEnv((workerId) => {
                    return Promise.resolve(f(workerId)).then(mapper);
                }, {
                    f,
                    mapper,
                });
            }, {
                mapper,
                depFunc,
                sf: sf.requireModule('@dcfjs/common/serializeFunction'),
            }),
            depFinalizer,
        ];
    }
    getNumPartitions() {
        return this._dependence.getNumPartitions();
    }
}
exports.MappedRDD = MappedRDD;
class UnionRDD extends RDD {
    constructor(context, dependences) {
        super(context);
        this._dependences = dependences;
    }
    async getFunc() {
        const partitionCounts = [];
        const rddFuncs = [];
        const rddFinalizers = [];
        for (let i = 0; i < this._dependences.length; i++) {
            const [numPartitions, func, finalizer] = await this._dependences[i].getFunc();
            partitionCounts.push(numPartitions);
            rddFuncs.push(func);
            if (finalizer) {
                rddFinalizers.push(finalizer);
            }
        }
        const numPartitions = partitionCounts.reduce((a, b) => a + b);
        return [
            numPartitions,
            sf.captureEnv((partitionId) => {
                for (let i = 0; i < partitionCounts.length; i++) {
                    if (partitionId < partitionCounts[i]) {
                        return rddFuncs[i](partitionId);
                    }
                    partitionId -= partitionCounts[i];
                }
                // `partitionId` should be less than totalPartitions.
                // so it should not reach here.
                throw new Error('Internal error.');
            }, {
                partitionCounts,
                rddFuncs,
            }),
            rddFinalizers.length === 0
                ? undefined
                : sf.captureEnv(async (ts) => {
                    for (const item of rddFinalizers) {
                        await item(ts);
                    }
                }, { rddFinalizers }),
        ];
    }
    getNumPartitions() {
        const partitionCounts = this._dependences.map((v) => v.getNumPartitions());
        if (partitionCounts.some((v) => v instanceof Promise)) {
            return Promise.all(partitionCounts).then((v) => v.reduce((a, b) => a + b));
        }
        return partitionCounts.reduce((a, b) => a + b);
    }
}
exports.UnionRDD = UnionRDD;
class CachedRDD extends RDD {
    constructor(context, dependence, storageType) {
        super(context);
        this._autoUnpersist = false;
        this._dependence = dependence;
        this._storageType = storageType;
    }
    // mark this cache auto unpersist after used again.
    // Used to clean up a temporary before last use.
    autoUnpersist() {
        this._autoUnpersist = true;
    }
    async unpersist() {
        const partitions = this._cachedPartitions;
        const storageType = this._storageType;
        if (partitions) {
            this._cachedPartitions = undefined;
            this._context.client.post('/exec', sf.serializeFunction(sf.captureEnv(((_, _1, tempStorageSession) => {
                for (const partition of partitions) {
                    tempStorageSession.release(storageType, partition);
                }
            }), {
                partitions,
                storageType,
            })));
        }
    }
    async getFunc() {
        const storageType = this._storageType;
        if (!this._cachedPartitions) {
            if (this._autoUnpersist) {
                // auto unpersist + not cached, use dependence directly.
                return await this._dependence.getFunc();
            }
            const [numPartitions, depFunc, depFinalizer,] = await this._dependence.getFunc();
            this._cachedNumPartition = numPartitions;
            // calc cached partitions.
            this._cachedPartitions = await this._context.execute(numPartitions, sf.captureEnv((partitionId, tempStorageSession) => {
                const f = depFunc(partitionId);
                return [
                    sf.captureEnv(async (workerId) => {
                        const data = await f(workerId);
                        const storage = sr.getTempStorage(storageType);
                        const key = await storage.generateKey();
                        storage.setItem(key, v8.serialize(data));
                        return key;
                    }, {
                        sr: sf.requireModule('@dcfjs/common/storageRegistry'),
                        v8: sf.requireModule('v8'),
                        f,
                        storageType,
                    }),
                    (key) => {
                        tempStorageSession.addRef(storageType, key);
                        return key;
                    },
                ];
            }, {
                sf: sf.requireModule('@dcfjs/common/serializeFunction'),
                storageType,
                depFunc,
            }), attachFinalizedFunc((arr) => arr, depFinalizer));
        }
        const autoUnpersist = this._autoUnpersist;
        const partitions = this._cachedPartitions;
        return [
            this._cachedNumPartition,
            sf.captureEnv((partitionId) => {
                const partition = partitions[partitionId];
                return sf.captureEnv(async () => {
                    const storage = sr.getTempStorage(storageType);
                    return v8.deserialize(autoUnpersist
                        ? await storage.getAndDeleteItem(partition)
                        : await storage.getItem(partition));
                }, {
                    autoUnpersist,
                    storageType,
                    partition,
                    sr: sf.requireModule('@dcfjs/common/storageRegistry'),
                    v8: sf.requireModule('v8'),
                });
            }, {
                autoUnpersist,
                partitions,
                storageType,
                sf: sf.requireModule('@dcfjs/common/serializeFunction'),
            }),
            autoUnpersist
                ? sf.captureEnv((ts) => {
                    for (const partition of partitions) {
                        ts.release(storageType, partition);
                    }
                }, {
                    partitions,
                    storageType,
                })
                : undefined,
        ];
    }
    getNumPartitions() {
        if (this._cachedPartitions) {
            return this._cachedNumPartition;
        }
        return this._dependence.getNumPartitions();
    }
}
exports.CachedRDD = CachedRDD;
async function getRepartitionFunc(context, dependence, numPartitions, partitionFunc) {
    const storageType = context.option.defaultRepartitionStorage;
    const [depPartitions, depFunc, depFinalizer] = dependence;
    const repartitionId = await context.client.post('/exec', sf.serializeFunction(sf.captureEnv((() => {
        return sr.getTempStorage(storageType).generateKey();
    }), {
        sr: sf.requireModule('@dcfjs/common/storageRegistry'),
        storageType,
    })));
    const pieces = await context.execute(depPartitions, sf.captureEnv((partitionId, tempStorageSession) => {
        const f = depFunc(partitionId);
        const pf = partitionFunc(partitionId);
        return [
            sf.captureEnv(async (workerId) => {
                const data = await f(workerId);
                const storage = sr.getTempStorage(storageType);
                const regrouped = pf(data);
                const ret = [];
                for (let i = 0; i < numPartitions; i++) {
                    if (!regrouped[i] || regrouped[i].length === 0) {
                        ret.push(null);
                        continue;
                    }
                    const key = `${repartitionId}-${workerId}-${i}`;
                    const buf = v8.serialize(regrouped[i]);
                    const len = Buffer.allocUnsafe(4);
                    len.writeUInt32LE(buf.length, 0);
                    await storage.appendItem(key, Buffer.concat([len, buf]));
                    ret.push(key);
                }
                return ret;
            }, {
                f,
                pf,
                numPartitions,
                repartitionId,
                storageType,
                sr: sf.requireModule('@dcfjs/common/storageRegistry'),
                v8: sf.requireModule('v8'),
            }),
            (keys) => {
                for (const key of keys) {
                    if (key) {
                        tempStorageSession.addRef(storageType, key);
                    }
                }
                return keys;
            },
        ];
    }, {
        sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        numPartitions,
        repartitionId,
        partitionFunc,
        storageType,
        depFunc,
    }), attachFinalizedFunc(sf.captureEnv((arr) => {
        // transposition matrix with new partition
        const ret = [];
        for (let i = 0; i < numPartitions; i++) {
            ret[i] = [];
            for (let j = 0; j < depPartitions; j++) {
                ret[i][j] = arr[j][i];
            }
        }
        return ret;
    }, {
        numPartitions,
        depPartitions,
    }), depFinalizer));
    return [
        numPartitions,
        sf.captureEnv((partitionId) => {
            const partPieces = pieces[partitionId];
            // TODO: use execution context to release references.
            return sf.captureEnv(async () => {
                const storage = sr.getTempStorage(storageType);
                const dataByKey = {};
                for (let i = 0; i < partPieces.length; i++) {
                    const key = partPieces[i];
                    if (key && !dataByKey[key]) {
                        const allBuf = await storage.getAndDeleteItem(key);
                        let position = 0;
                        const tmp = [];
                        while (position < allBuf.length) {
                            const len = allBuf.readUInt32LE(position);
                            position += 4;
                            const buf = allBuf.slice(position, position + len);
                            position += len;
                            tmp.push(v8.deserialize(buf));
                        }
                        dataByKey[key] = tmp;
                    }
                }
                const parts = [];
                for (let i = 0; i < partPieces.length; i++) {
                    const key = partPieces[i];
                    if (key) {
                        parts.push(dataByKey[key].shift());
                    }
                }
                return ah.concatArrays(parts);
            }, {
                partPieces,
                storageType,
                depPartitions,
                sr: sf.requireModule('@dcfjs/common/storageRegistry'),
                v8: sf.requireModule('v8'),
                ah: sf.requireModule('@dcfjs/common/arrayHelpers'),
            });
        }, {
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
            storageType,
            numPartitions,
            depPartitions,
            pieces,
        }),
        sf.captureEnv((ts) => {
            for (const partition of pieces) {
                for (const piece of partition) {
                    if (piece) {
                        ts.release(storageType, piece);
                    }
                }
            }
        }, {
            pieces,
            storageType,
        }),
    ];
}
class RepartitionRDD extends RDD {
    constructor(context, dependence, numPartition, partitionFunc) {
        super(context);
        this._numPartition = numPartition;
        this._dependence = dependence;
        this._partitionFunc = partitionFunc;
    }
    async getFunc() {
        return getRepartitionFunc(this._context, await this._dependence.getFunc(), this._numPartition, this._partitionFunc);
    }
    getNumPartitions() {
        return this._numPartition;
    }
}
exports.RepartitionRDD = RepartitionRDD;
class SortedRDD extends RDD {
    constructor(context, dependence, numPartition, keyFunc, ascending) {
        super(context);
        this._numPartition = numPartition;
        this._dependence = dependence;
        this._numPartition = numPartition;
        this._keyFunc = keyFunc;
        this._ascending = ascending;
    }
    async getFunc() {
        let cachedDependence;
        const storageType = this._context.option.defaultRepartitionStorage;
        // Step1: cache if needed.
        if (this._dependence instanceof CachedRDD) {
            cachedDependence = this._dependence;
        }
        else {
            cachedDependence = this._dependence.persist(storageType);
        }
        // Step 2:
        // sample with 1/n fraction where n is origin partition count
        // so we get a sample count near to a single partition.
        // Sample contains partitionIndex & localIndex to avoid performance inssue
        // when dealing with too many same values.
        const [originPartitions, originFunc, originFinalizer,] = await cachedDependence.getFunc();
        const keyFunc = this._keyFunc;
        const samples = await this._context.execute(originPartitions, sf.captureEnv((partitionId) => {
            const f = originFunc(partitionId);
            return sf.captureEnv(async (workerId) => {
                const v = await f(workerId);
                const ret = [];
                for (let i = 0; i < v.length; i++) {
                    if (Math.random() * originPartitions < 1) {
                        ret.push([keyFunc(v[i]), partitionId, i]);
                    }
                }
                return ret;
            }, { f, partitionId, keyFunc, originPartitions });
        }, {
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
            originPartitions,
            originFunc,
            keyFunc,
        }), attachFinalizedFunc(sf.captureEnv((arr) => {
            return ah.concatArrays(arr);
        }, {
            ah: sf.requireModule('@dcfjs/common/arrayHelpers'),
        }), originFinalizer));
        const ascending = this._ascending;
        // Step 3: sort samples, and get seperate points.
        samples.sort((a, b) => {
            if (a[0] !== b[0]) {
                return (a[0] < b[0] ? -1 : 1) * (ascending ? 1 : -1);
            }
            if (a[1] !== b[1]) {
                return a[1] < b[1] ? -1 : 1;
            }
            return a[2] < b[2] ? -1 : 1;
        });
        // get n-1 points in samples.
        const numPartitions = this._numPartition;
        const points = [];
        let p = samples.length / numPartitions;
        for (let i = 1; i < numPartitions; i++) {
            let idx = Math.floor(p * i);
            if (idx < samples.length) {
                points.push(samples[idx]);
            }
        }
        // Step 4: Repartition by points.
        if (this._dependence !== cachedDependence) {
            // remove persisted data after sorted.
            cachedDependence.autoUnpersist();
        }
        const [_, repartitionedFunc, finalizer] = await getRepartitionFunc(this._context, [originPartitions, originFunc, originFinalizer], numPartitions, sf.captureEnv((partitionId) => {
            return sf.captureEnv((v) => {
                const ret = [];
                for (let i = 0; i < numPartitions; i++) {
                    ret.push([]);
                }
                const tmp = v.map((item, i) => [keyFunc(item), i]);
                tmp.sort((a, b) => {
                    if (a[0] !== b[0]) {
                        return (a[0] < b[0] ? -1 : 1) * (ascending ? 1 : -1);
                    }
                    return a[1] < b[1] ? -1 : 1;
                });
                let index = 0;
                for (const [key, i] of tmp) {
                    // compare with points
                    for (; index < points.length;) {
                        if (ascending ? key < points[index][0] : key > points[index][0]) {
                            break;
                        }
                        if (ascending ? key > points[index][0] : key < points[index][0]) {
                            index++;
                            continue;
                        }
                        if (partitionId < points[index][1]) {
                            break;
                        }
                        if (partitionId > points[index][1]) {
                            index++;
                            continue;
                        }
                        if (i < points[index][2]) {
                            break;
                        }
                        index++;
                    }
                    ret[index].push([v[i], key, partitionId, i]);
                }
                return ret;
            }, {
                keyFunc,
                numPartitions,
                partitionId,
                points,
                ascending,
            });
        }, {
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
            keyFunc,
            numPartitions,
            points,
            ascending,
        }));
        // Step 5: Sort in partition.
        // Maybe reverse (if ascending == false).
        return [
            numPartitions,
            sf.captureEnv((partitionId) => {
                const f = repartitionedFunc(partitionId);
                return sf.captureEnv(async (workerId) => {
                    const tmp = await f(workerId);
                    tmp.sort((a, b) => {
                        if (a[1] !== b[1]) {
                            return (a[1] < b[1] ? -1 : 1) * (ascending ? 1 : -1);
                        }
                        if (a[2] !== b[2]) {
                            return a[2] < b[2] ? -1 : 1;
                        }
                        return a[3] < b[3] ? -1 : 1;
                    });
                    return tmp.map((item) => item[0]);
                }, {
                    f,
                    ascending,
                });
            }, {
                sf: sf.requireModule('@dcfjs/common/serializeFunction'),
                numPartitions,
                repartitionedFunc,
                ascending,
            }),
            finalizer,
        ];
    }
    getNumPartitions() {
        return this._numPartition;
    }
}
exports.SortedRDD = SortedRDD;
class CoalesceRDD extends RDD {
    constructor(context, dependence, numPartition) {
        super(context);
        this._numPartition = numPartition;
        this._dependence = dependence;
    }
    async getFunc() {
        const [originPartitions, originFunc, originFinalizer,] = await this._dependence.getFunc();
        const numPartitions = this._numPartition;
        const partitionArgs = [];
        let last = [];
        partitionArgs.push([0, last]);
        const rate = originPartitions / numPartitions;
        let counter = 0;
        for (let i = 0; i < numPartitions - 1; i++) {
            counter += rate;
            while (counter >= 1) {
                counter -= 1;
                last = [];
                partitionArgs.push([i, last]);
            }
            last.push(counter);
        }
        // manually add last partition to avoid precsion loss.
        while (partitionArgs.length < originPartitions) {
            partitionArgs.push([numPartitions - 1, []]);
        }
        return getRepartitionFunc(this._context, [originPartitions, originFunc, originFinalizer], numPartitions, sf.captureEnv((partitionId) => {
            const arg = partitionArgs[partitionId];
            return sf.captureEnv((data) => {
                const ret = [];
                let lastIndex = 0;
                for (let i = 0; i < arg[0]; i++) {
                    ret.push([]);
                }
                for (const rate of arg[1]) {
                    const nextIndex = Math.floor(data.length * rate);
                    ret.push(data.slice(lastIndex, nextIndex));
                    lastIndex = nextIndex;
                }
                ret.push(data.slice(lastIndex));
                return ret;
            }, { numPartitions, arg });
        }, {
            numPartitions,
            partitionArgs,
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        }));
    }
    getNumPartitions() {
        return this._numPartition;
    }
}
exports.CoalesceRDD = CoalesceRDD;
class FileLoaderRDD extends RDD {
    constructor(context, fileListFunc, loaderFunc) {
        super(context);
        this._fileListFunc = fileListFunc;
        this._loaderFunc = loaderFunc;
    }
    async _getFileList() {
        if (!this._fileList) {
            this._fileList = await this._fileListFunc();
        }
    }
    async getFunc() {
        await this._getFileList();
        const fileList = this._fileList;
        const loaderFunc = this._loaderFunc;
        return [
            fileList.length,
            sf.captureEnv((partitionId) => {
                const fileName = fileList[partitionId];
                return sf.captureEnv(async () => {
                    return [
                        [fileName, await loaderFunc(fileName)],
                    ];
                }, {
                    fileName,
                    loaderFunc,
                });
            }, {
                sf: sf.requireModule('@dcfjs/common/serializeFunction'),
                loaderFunc,
                fileList,
            }),
            undefined,
        ];
    }
    async getNumPartitions() {
        this._getFileList();
        return this._fileList.length;
    }
}
exports.FileLoaderRDD = FileLoaderRDD;
