"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("@dcfjs/common/noCaptureEnv");
const sf = require("@dcfjs/common/serializeFunction");
const fs = require("fs");
const path = require("path");
const os = require("os");
const url = require("url");
function walkSync(dirPath, filelist, recursive = false) {
    if (!dirPath.endsWith('/')) {
        dirPath = dirPath + '/';
    }
    const files = fs.readdirSync(dirPath);
    files.forEach(function (file) {
        if (recursive) {
            if (fs.statSync(dirPath + file).isDirectory()) {
                walkSync(dirPath + file + '/', filelist, recursive);
            }
            else {
                filelist.push(dirPath + file);
            }
        }
        else {
            if (!fs.statSync(dirPath + file).isDirectory()) {
                filelist.push(dirPath + file);
            }
        }
    });
}
const capturedWalkSync = sf.captureEnv(walkSync, {
    fs: sf.requireModule('fs'),
});
const recursiveMkdirSync = sf.captureEnv(function (dirPath) {
    if (!dirPath.startsWith('/')) {
        throw new Error('only absoulte paths are supported');
    }
    if (fs.existsSync(dirPath)) {
        return;
    }
    let curPath = '/';
    const tokens = dirPath.split('/').filter(v => v.length > 0);
    for (const token of tokens) {
        curPath = curPath + token + '/';
        if (!fs.existsSync(curPath)) {
            fs.mkdirSync(curPath);
        }
    }
}, {
    fs: sf.requireModule('fs'),
});
// remove all content in a directory except itself.
const rmdirs = sf.captureEnv(function (basePath) {
    function work(basePath) {
        const stat = fs.lstatSync(basePath);
        if (stat.isDirectory()) {
            const contents = fs.readdirSync(basePath);
            for (const file of contents) {
                work(path.resolve(basePath, file));
            }
            fs.rmdirSync(basePath);
        }
        else {
            fs.unlinkSync(basePath);
        }
    }
    const contents = fs.readdirSync(basePath);
    for (const file of contents) {
        work(path.resolve(basePath, file));
    }
}, {
    path: sf.requireModule('path'),
    fs: sf.requireModule('fs'),
});
const solvePath = sf.captureEnv((baseUri, ...filenames) => {
    const uri = url.parse(baseUri);
    if (uri.protocol) {
        // absolute path
        const ret = uri.path;
        if (os.platform() === 'win32') {
            return path.resolve(ret.substr(1), ...filenames);
        }
        return path.resolve(ret, ...filenames);
    }
    else {
        // maybe relative
        return path.resolve(process.cwd(), baseUri, ...filenames);
    }
}, {
    os: sf.requireModule('os'),
    path: sf.requireModule('path'),
    url: sf.requireModule('url'),
});
class SharedFsLoader {
    constructor() {
        this.listFiles = sf.captureEnv((baseUri, recursive) => {
            const fileList = [];
            capturedWalkSync(baseUri, fileList, recursive);
            return fileList;
        }, {
            capturedWalkSync,
        });
        this.loadFile = sf.captureEnv((baseUri, filename) => {
            const fn = solvePath(baseUri, filename);
            return fs.readFileSync(fn);
        }, {
            fs: sf.requireModule('fs'),
            path: sf.requireModule('path'),
            solvePath,
        });
        this.initSaveProgress = sf.captureEnv((baseUri, overwrite) => {
            const basePath = solvePath(baseUri);
            if (!fs.existsSync(basePath)) {
                recursiveMkdirSync(basePath);
            }
            else {
                if (!overwrite) {
                    throw new Error(`${basePath} already exists, consider use overwrite=true?`);
                }
                rmdirs(basePath);
            }
            // Create a writing mark.
            fs.writeFileSync(path.resolve(basePath, '.writing'), Buffer.alloc(0));
        }, {
            fs: sf.requireModule('fs'),
            path: sf.requireModule('path'),
            recursiveMkdirSync,
            solvePath,
        });
        this.saveFile = sf.captureEnv((baseUri, filename, data) => {
            const path = solvePath(baseUri, filename);
            fs.writeFileSync(path, data);
        }, {
            fs: sf.requireModule('fs'),
            solvePath,
        });
        this.markSaveSuccess = sf.captureEnv((baseUri) => {
            const basePath = solvePath(baseUri);
            fs.renameSync(path.resolve(basePath, '.writing'), path.resolve(basePath, '.success'));
        }, {
            fs: sf.requireModule('fs'),
            path: sf.requireModule('path'),
            solvePath,
        });
    }
    canHandleUri(baseUri) {
        const uri = url.parse(baseUri);
        if (uri.protocol && uri.protocol !== 'file:') {
            return false;
        }
        return true;
    }
}
exports.SharedFsLoader = SharedFsLoader;
